import hashlib
import inspect
import json
import logging
import os
import re
import traceback
from collections import defaultdict
from typing import Any, TYPE_CHECKING, Union

from Levenshtein import ratio

from baserun import Baserun
from baserun.grpc import (
    get_or_create_submission_service,
    get_or_create_async_submission_service,
)
from baserun.helpers import memoize_for_time
from baserun.v1.baserun_pb2 import (
    Template,
    TemplateVersion,
    SubmitTemplateVersionRequest,
    SubmitTemplateVersionResponse,
    GetTemplatesRequest,
    GetTemplatesResponse,
)

if TYPE_CHECKING:
    # Just for type annotations for Langchain. Since langchain is an optional dependency we have this garbage
    try:
        from langchain.tools import Tool
    except ImportError:
        Tool = Any

logger = logging.getLogger(__name__)


@memoize_for_time(os.environ.get("BASERUN_CACHE_INTERVAL", 600))
def get_templates(environment: str = None) -> dict[str, Template]:
    if not Baserun.templates:
        Baserun.templates = {}

    try:
        request = GetTemplatesRequest(environment=environment or Baserun.environment)
        response: GetTemplatesResponse = get_or_create_submission_service().GetTemplates(request)
        for template in response.templates:
            Baserun.templates[template.name] = template

    except BaseException as e:
        logger.error(f"Could not fetch templates from Baserun. Using {len(Baserun.templates.keys())} cached templates")
        logger.info(traceback.format_exception(e))

    return Baserun.templates


def get_template(name: str) -> Template:
    templates = get_templates()
    template = templates.get(name)
    if not template:
        logger.info(f"Attempted to get template {name} but no template with that name exists")
        return None

    return template


def most_similar_templates(formatted_str: str):
    """Given a `templates` list, will return the templates sorted by similarity to the `formatted_str` arg"""
    if not Baserun.templates:
        Baserun.templates = {}
        return []

    similarity_scores = []

    for template in Baserun.templates.values():
        similarity_scores.append(
            (template.active_version, ratio(formatted_str, template.active_version.template_string))
        )

    # Sort the templates by similarity ratio, in descending order
    sorted_results = sorted(similarity_scores, key=lambda x: x[1], reverse=True)

    # Extract only the TemplateVersion objects from the sorted list
    sorted_templates = [template for template, score in sorted_results if score > 0.5]

    return sorted_templates


def best_guess_template_parameters(prompt: str, template_version: TemplateVersion) -> dict[str, Any]:
    """Given a prompt and a TemplateVersion will attempt to find the parameters that were used to generate the prompt
    Major caveats apply: this only works if the template was registered and formatted using `format_prompt` and/or
    the parameters were registered with `capture_parameters`. It's also pretty fragile because it's based on simple
    string comparison.

    The correct version of this is probably some span context propagation that I can't figure out right now.
    """
    used_parameter_list = Baserun.used_template_parameters.get(template_version.id, [])
    for used_parameters in used_parameter_list:
        # TODO: This only works for string parameters
        match = all(parameter in prompt for parameter in used_parameters.values() if isinstance(parameter, str))
        if match:
            return used_parameters

    return {}


def is_from_jinja2_template(template_str: str, formatted_str: str):
    # Use regular expressions to extract non-variable content
    template_content = re.sub(r"{{.*?}}", "", template_str)

    # Check if all characters in template_content are present in the formatted string
    return all(char in formatted_str for char in template_content)


def extract_strings_in_braces(input_string):
    # Matches both single and double braces around the variable name
    pattern = r"\{{1,2}([^}]*)\}{1,2}"
    matches = re.findall(pattern, input_string)
    return matches


def get_template_type_enum(template_type: str = None):
    template_type = template_type or "unknown"
    if template_type == Template.TEMPLATE_TYPE_JINJA2 or template_type.lower().startswith("jinja"):
        template_type_enum = Template.TEMPLATE_TYPE_JINJA2
    else:
        template_type_enum = Template.TEMPLATE_TYPE_FORMATTED_STRING

    return template_type_enum


def apply_template(template_string: str, parameters: dict[str, Any], template_type_enum) -> str:
    if template_type_enum == Template.TEMPLATE_TYPE_JINJA2:
        try:
            # noinspection PyUnresolvedReferences
            from jinja2 import Template as JinjaTemplate

            template = JinjaTemplate(template_string)
            return template.render(parameters)
        except ImportError:
            logger.warning("Cannot render Jinja2 template as jinja2 package is not installed")
            # TODO: Is this OK? should we raise? or return blank string?
            return template_string

    return template_string.format(**parameters)


def capture_parameters(template_version_id: str, parameters: dict[str, Any]):
    if not Baserun.used_template_parameters:
        Baserun.used_template_parameters = defaultdict(list)

    Baserun.used_template_parameters[template_version_id].append(parameters)


def create_langchain_template(
    template_string: str,
    parameters: dict[str, Any] = None,
    template_name: str = None,
    template_tag: str = None,
    template_type: str = None,
    tools: list[Union["Tool", Any]] = None,
):
    from langchain.prompts import PromptTemplate

    parameters = parameters or {}
    input_variables = list(parameters.keys())
    template_type = template_type or "Formatted String"
    tools = tools or []

    if not template_name:
        caller = inspect.stack()[1].function
        template_name = f"{caller}_template"

    if template_type == Template.TEMPLATE_TYPE_JINJA2 or template_type.lower().startswith("jinja"):
        langchain_template = PromptTemplate(
            template=template_string,
            input_variables=input_variables,
            template_format="jinja2",
            tools=tools,
        )

    else:
        langchain_template = PromptTemplate(template=template_string, input_variables=input_variables, tools=tools)

    template_type_enum = get_template_type_enum(template_type)
    # Support only strings for now
    parameter_definition = {var: "string" for var in input_variables}
    template_version = register_template(
        template_string=template_string,
        template_name=template_name,
        template_type=template_type_enum,
        template_tag=template_tag,
        parameter_definition=parameter_definition,
    )

    capture_parameters(template_version.id, parameters)

    return langchain_template


def format_prompt(
    template_string: str,
    parameters: dict[str, Any],
    template_name: str = None,
    template_tag: str = None,
    template_type: str = None,
    parameter_definition: dict[str, Any] = None,
):
    template_type_enum = get_template_type_enum(template_type)
    try:
        template_version = register_template(
            template_string=template_string,
            template_name=template_name,
            template_type=template_type_enum,
            template_tag=template_tag,
            parameter_definition=parameter_definition,
        )
    except BaseException as e:
        logger.warning(f"Could not register template: {e}")
        return

    capture_parameters(template_version.id, parameters)
    return apply_template(template_string, parameters, template_type_enum)


async def aformat_prompt(
    template_string: str,
    parameters: dict[str, Any],
    template_name: str = None,
    template_tag: str = None,
    template_type: str = None,
    parameter_definition: dict[str, Any] = None,
):
    template_type_enum = get_template_type_enum(template_type)
    try:
        version = await aregister_template(
            template_string=template_string,
            template_name=template_name,
            template_type=template_type_enum,
            template_tag=template_tag,
            parameter_definition=parameter_definition,
        )
    except BaseException as e:
        logger.warning(f"Could not register template: {e}")
        return

    capture_parameters(version.id, parameters)
    return apply_template(template_string, parameters, template_type_enum)


def construct_template_version(
    template_string: str,
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
    parameter_definition: dict[str, Any] = None,
) -> TemplateVersion:
    if not parameter_definition:
        parameters = extract_strings_in_braces(template_string)
        parameter_definition = {p: "string" for p in parameters}

    # Automatically generate a name based on the template's contents
    if not template_name:
        template_name = hashlib.sha256(f"{template_string}{parameter_definition}".encode()).hexdigest()[:5]

    if not template_tag:
        template_tag = hashlib.sha256(template_string.encode()).hexdigest()[:5]

    template = Template(name=template_name, template_type=template_type)
    version = TemplateVersion(
        template=template,
        parameter_definition=json.dumps(parameter_definition),
        template_string=template_string,
        tag=template_tag,
    )

    return version


def register_template(
    template_string: str,
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
    parameter_definition: dict[str, Any] = None,
) -> Template:
    from baserun import Baserun

    if not Baserun.templates:
        Baserun.templates = {}

    if template := Baserun.templates.get(template_name):
        return template

    version = construct_template_version(
        template_string=template_string,
        template_name=template_name,
        template_tag=template_tag,
        template_type=template_type,
        parameter_definition=parameter_definition,
    )

    request = SubmitTemplateVersionRequest(template_version=version, environment=Baserun.environment)
    response: SubmitTemplateVersionResponse = get_or_create_submission_service().SubmitTemplateVersion(request)

    template = response.template
    if template.name not in Baserun.templates:
        Baserun.templates[template.name] = response.template

    return response.template


async def aregister_template(
    template_string: str,
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
    parameter_definition: dict[str, Any] = None,
) -> Template:
    from baserun import Baserun

    if not Baserun.templates:
        Baserun.templates = {}

    if template := Baserun.templates.get(template_name):
        return template

    version = construct_template_version(
        template_string=template_string,
        template_name=template_name,
        template_tag=template_tag,
        template_type=template_type,
        parameter_definition=parameter_definition,
    )

    request = SubmitTemplateVersionRequest(template_version=version, environment=Baserun.environment)
    response: SubmitTemplateVersionResponse = await get_or_create_async_submission_service().SubmitTemplateVersion(
        request
    )

    template = response.template
    if template.name not in Baserun.templates:
        Baserun.templates[template.name] = template

    return template



import collections.abc
import json
import logging
from abc import abstractmethod
from datetime import datetime
from typing import Callable
from uuid import UUID

import httpx
from openai._base_client import BaseClient
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice
from opentelemetry.instrumentation.instrumentor import (
    BaseInstrumentor as OTelInstrumentor,
)
from opentelemetry.sdk.trace import _Span

from baserun.v1.baserun_pb2 import Span, Message, Status, SubmitSpanRequest, ToolCall, ToolFunction

logger = logging.getLogger(__name__)


def spy_on_process_response_data(original_method):
    from baserun import Baserun

    def wrapper(self, *args, **kwargs):
        result: ChatCompletionChunk = original_method(self, *args, **kwargs)
        if not isinstance(result, ChatCompletionChunk):
            return result

        try:
            response = kwargs.get("response")

            if not hasattr(response, "deltas"):
                setattr(response, "deltas", [])

            response.deltas.append(result)
            first_delta = result.choices[0].delta

            # Stream has ended, compile the deltas and submit
            if first_delta.content is None and first_delta.function_call is None and first_delta.tool_calls is None:
                content = ""
                function_name = ""
                function_args = ""
                tool_calls = []
                for delta in response.deltas:
                    if new_content := delta.choices[0].delta.content:
                        content += new_content

                    if new_function := delta.choices[0].delta.function_call:
                        if new_function.name:
                            function_name += new_function.name
                        if new_function.arguments:
                            function_args += new_function.arguments

                    if new_tool_calls := delta.choices[0].delta.tool_calls:
                        for new_tool_call in new_tool_calls:
                            while len(tool_calls) <= new_tool_call.index:
                                tool_calls.append(ToolCall(function=ToolFunction()))

                            tool_call = tool_calls[new_tool_call.index]

                            if new_tool_call.id:
                                tool_call.id = new_tool_call.id

                            if new_tool_call.type:
                                tool_call.type += new_tool_call.type

                            if new_tool_call.function.name:
                                tool_call.function.name += new_tool_call.function.name

                            if new_tool_call.function.arguments:
                                tool_call.function.arguments += new_tool_call.function.arguments

                first_delta = response.deltas[0]

                completion_message = Message(
                    # Role is set on the first delta
                    role=first_delta.choices[0].delta.role,
                    # Content is the aggregate of all deltas
                    content=content,
                    # Finish reason is on the last delta
                    finish_reason=result.choices[0].finish_reason,
                )
                if function_name:
                    completion_message.function_call = json.dumps({"name": function_name, "arguments": function_args})

                if tool_calls:
                    for tool_call in tool_calls:
                        completion_message.tool_calls.append(tool_call)

                span_request = response._span_request
                span = span_request.span
                span.completions.append(completion_message)
                span.completion_id = result.id
                span.model = result.model
                # 1 token per delta, minus 2 (first = empty content, last = None content)
                span.completion_tokens = len(response.deltas) - 2
                # TODO: How does one count prompt tokens here?
                span.prompt_tokens = 0
                span.total_tokens = span.completion_tokens + span.prompt_tokens

                # TODO: Templates

                Baserun.exporter_queue.put(span_request)
        except BaseException as e:
            logger.warning(f"Failed to collect span for Baserun: {e}")
            pass

        return result

    return wrapper


def compile_tool_calls(choice: Choice) -> list[ToolCall]:
    calls = []
    if not choice.message.tool_calls:
        return calls

    for call in choice.message.tool_calls:
        calls.append(
            ToolCall(
                id=call.id,
                type=call.type,
                function=ToolFunction(name=call.function.name, arguments=call.function.arguments),
            )
        )

    return calls


def spy_on_process_response(original_method):
    def wrapper(self, *args, **kwargs):
        from baserun import Baserun
        import openai

        start_time = datetime.utcnow()
        result: ChatCompletion = original_method(self, *args, **kwargs)
        end_time = datetime.utcnow()
        response: httpx.Response = kwargs.get("response")
        if response:
            try:
                request: httpx.Request = response.request
                parsed_request = json.loads(request.content)
                current_run = Baserun.current_run()
                if not current_run:
                    # TODO handle this case
                    return result

                prompt_messages = [
                    Message(
                        role=message.get("role"),
                        content=message.get("content"),
                    )
                    for message in parsed_request.get("messages", [])
                ]

                if hasattr(result, "choices"):
                    completion_messages = [
                        Message(
                            role=choice.message.role,
                            content=choice.message.content,
                            finish_reason=choice.finish_reason,
                            function_call=choice.message.function_call,
                            tool_calls=compile_tool_calls(choice),
                            system_fingerprint=result.system_fingerprint,
                        )
                        for choice in result.choices
                    ]
                else:
                    # Streaming response, will set the completion data later
                    completion_messages = []

                x_request_id = response.headers.get("x-request-id")
                span = Span(
                    # TODO: Message for non-200
                    status=Status(message="ok", code=response.status_code),
                    prompt_messages=prompt_messages,
                    completions=completion_messages,
                )

                if tools := parsed_request.get("tools"):
                    span.tools = json.dumps(tools)

                if functions := parsed_request.get("functions"):
                    span.functions = json.dumps(functions)

                span.run_id = current_run.run_id
                span.trace_id = UUID(current_run.run_id).bytes
                span.span_id = int.from_bytes(bytes.fromhex(x_request_id[:8]), "big")
                span.name = "openai.chat"
                span.vendor = "openai"
                span.request_type = "chat"
                span.start_time.FromDatetime(start_time)
                span.end_time.FromDatetime(end_time)

                span.api_type = openai.api_type or "open_ai"
                span.api_base = openai.base_url or "https://api.openai.com/v1"
                span.x_request_id = x_request_id
                span.stream = parsed_request.get("stream", False)

                if max_tokens := parsed_request.get("max_tokens"):
                    span.max_tokens = max_tokens
                if temperature := parsed_request.get("temperature"):
                    span.temperature = temperature
                if top_p := parsed_request.get("top_p"):
                    span.top_p = top_p
                if top_k := parsed_request.get("top_k"):
                    span.top_k = top_k
                if frequency_penalty := parsed_request.get("frequency_penalty"):
                    span.frequency_penalty = frequency_penalty
                if presence_penalty := parsed_request.get("presence_penalty"):
                    span.presence_penalty = presence_penalty
                if n := parsed_request.get("n"):
                    span.n = n
                if logit_bias := parsed_request.get("logit_bias"):
                    span.logit_bias = logit_bias
                if logprobs := parsed_request.get("logprobs"):
                    span.logprobs = logprobs
                if stop := kwargs.get("stop"):
                    if isinstance(stop, str):
                        span.stop = [stop]
                    else:
                        span.stop = stop
                if echo := parsed_request.get("echo"):
                    span.echo = echo
                if suffix := parsed_request.get("suffix"):
                    span.suffix = suffix
                if best_of := parsed_request.get("best_of"):
                    span.best_of = best_of
                if user := parsed_request.get("user"):
                    span.user = user
                if function_call := parsed_request.get("function_call"):
                    span.function_call = json.dumps(function_call)
                if tool_choice := parsed_request.get("tool_choice"):
                    span.tool_choice = json.dumps(tool_choice)

                # Non-streaming
                if hasattr(result, "model"):
                    span.model = result.model
                    span.total_tokens = result.usage.total_tokens
                    span.completion_tokens = result.usage.completion_tokens
                    span.prompt_tokens = result.usage.prompt_tokens
                    span.completion_id = result.id

                    span_request = SubmitSpanRequest(span=span, run=current_run)
                    # TODO: Templates
                    Baserun.exporter_queue.put(span_request)
                    return result
                # Streaming- will get submitted in `spy_on_process_response_data` after streaming finishes
                else:
                    span_request = SubmitSpanRequest(span=span, run=current_run)
                    setattr(response, "_span_request", span_request)

            except BaseException as e:
                logger.warning(f"Failed to collect span for Baserun: {e}")
                pass

        return result

    return wrapper


original_methods = {}


def instrument():
    global original_methods
    if original_methods:
        return

    original_methods = {
        "_process_response": BaseClient._process_response,
        "_process_response_data": BaseClient._process_response_data,
    }
    BaseClient._process_response = spy_on_process_response(BaseClient._process_response)
    BaseClient._process_response_data = spy_on_process_response_data(BaseClient._process_response_data)


def uninstrument():
    BaseClient._process_response = original_methods["_process_response"]
    BaseClient._process_response_data = original_methods["_process_response_data"]


class BaseInstrumentor(OTelInstrumentor):
    original_methods: dict[str, Callable] = None

    @staticmethod
    @abstractmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        pass

    @staticmethod
    @abstractmethod
    def async_generator_wrapper(original_generator: collections.abc.AsyncIterator, span: _Span):
        pass

    def _instrument(self, **kwargs):
        self.original_methods = {
            "_process_response": BaseClient._process_response,
            "_process_response_data": BaseClient._process_response_data,
        }
        BaseClient._process_response = spy_on_process_response(BaseClient._process_response)
        BaseClient._process_response_data = spy_on_process_response_data(BaseClient._process_response_data)

    def _uninstrument(self, **kwargs):
        BaseClient._process_response = self.original_methods["_process_response"]
        BaseClient._process_response_data = self.original_methods["_process_response_data"]
